@@ -3,34 +3,35 @@
 
 from __future__ import absolute_import, division, print_function, unicode_literals
 
-import os, sys, shutil, argparse, subprocess
+import os, sys, shutil, argparse, subprocess, unittest, contextlib
 import pexpect, pexpect.replwrap
-from tempfile import TemporaryFile, mkdtemp
+from tempfile import TemporaryFile, NamedTemporaryFile, mkdtemp
+from io import StringIO
 
-TEST_DIR = os.path.abspath(os.path.dirname(__file__))
-BASE_DIR = os.path.dirname(TEST_DIR)
-sys.path.insert(0, BASE_DIR)
+TEST_DIR = os.path.abspath(os.path.dirname(__file__))  # noqa
+BASE_DIR = os.path.dirname(TEST_DIR)  # noqa
+sys.path.insert(0, BASE_DIR)  # noqa
 
 from argparse import ArgumentParser, SUPPRESS
+import argcomplete
 from argcomplete import (
     autocomplete,
     CompletionFinder,
     split_line,
     ExclusiveCompletionFinder,
+    _check_module,
+    shellcode,
+    warn
 )
-from argcomplete.completers import FilesCompleter, DirectoriesCompleter
-from argcomplete.compat import USING_PYTHON2, str, sys_encoding, ensure_str, ensure_bytes
-
-if sys.version_info >= (2, 7):
-    import unittest
-else:
-    import unittest2 as unittest
+from argcomplete.completers import FilesCompleter, DirectoriesCompleter, SuppressCompleter
 
 IFS = "\013"
-COMP_WORDBREAKS = " \t\n\"'><=;&(:"
+COMP_WORDBREAKS = " \t\n\"'><=;|&(:"
 
 BASH_VERSION = subprocess.check_output(['bash', '-c', 'echo $BASH_VERSION']).decode()
 BASH_MAJOR_VERSION = int(BASH_VERSION.split('.')[0])
+FISH_VERSION_STR = subprocess.check_output(['fish', '-c', 'echo -n $version']).decode()
+FISH_VERSION_TUPLE = tuple(int(x) for x in FISH_VERSION_STR.split('.'))
 
 
 class TempDir(object):
@@ -61,7 +62,7 @@ class TestArgcomplete(unittest.TestCase):
     def setUp(self):
         self._os_environ = os.environ
         os.environ = os.environ.copy()
-        os.environ["_ARGCOMPLETE"] = "yes"
+        os.environ["_ARGCOMPLETE"] = "1"
         os.environ["_ARC_DEBUG"] = "yes"
         os.environ["IFS"] = IFS
         os.environ["_ARGCOMPLETE_COMP_WORDBREAKS"] = COMP_WORDBREAKS
@@ -70,18 +71,17 @@ class TestArgcomplete(unittest.TestCase):
         os.environ = self._os_environ
 
     def run_completer(self, parser, command, point=None, completer=autocomplete, **kwargs):
-        command = ensure_str(command)
-        
         if point is None:
-            # Adjust point for wide chars
-            point = str(len(command.encode(sys_encoding)))
-        with TemporaryFile() as t:
-            os.environ["COMP_LINE"] = ensure_bytes(command) if USING_PYTHON2 else command
+            point = str(len(command))
+        with TemporaryFile(mode="w+") as t:
+            os.environ["COMP_LINE"] = command
             os.environ["COMP_POINT"] = point
-            self.assertRaises(SystemExit, completer, parser, output_stream=t,
-                              exit_method=sys.exit, **kwargs)
+            with self.assertRaises(SystemExit) as cm:
+                completer(parser, output_stream=t, exit_method=sys.exit, **kwargs)
+            if cm.exception.code != 0:
+                raise Exception("Unexpected exit code %d" % cm.exception.code)
             t.seek(0)
-            return t.read().decode(sys_encoding).split(IFS)
+            return t.read().split(IFS)
 
     def test_basic_completion(self):
         p = ArgumentParser()
@@ -103,7 +103,7 @@ class TestArgcomplete(unittest.TestCase):
     def test_choices(self):
         def make_parser():
             parser = ArgumentParser()
-            parser.add_argument("--ship", choices=["submarine", b"speedboat"])
+            parser.add_argument("--ship", choices=["submarine", "speedboat"])
             return parser
 
         expected_outputs = (
@@ -111,7 +111,10 @@ class TestArgcomplete(unittest.TestCase):
             ("prog --shi", ["--ship "]),
             ("prog --ship ", ["submarine", "speedboat"]),
             ("prog --ship s", ["submarine", "speedboat"]),
-            ("prog --ship su", ["submarine "])
+            ("prog --ship su", ["submarine "]),
+            ("prog --ship=", ["submarine", "speedboat"]),
+            ("prog --ship=s", ["submarine", "speedboat"]),
+            ("prog --ship=su", ["submarine "]),
         )
 
         for cmd, output in expected_outputs:
@@ -157,6 +160,30 @@ class TestArgcomplete(unittest.TestCase):
         for cmd, output in expected_outputs:
             self.assertEqual(set(self.run_completer(make_parser(), cmd, print_suppressed=True)), set(output))
 
+    def test_suppress_completer(self):
+        def make_parser():
+            parser = ArgumentParser()
+            parser.add_argument("--foo")
+            arg = parser.add_argument("--bar")
+            arg.completer = SuppressCompleter()
+            return parser
+
+        expected_outputs = (
+            ("prog ", ["--foo", "-h", "--help"]),
+            ("prog --b", [""])
+        )
+
+        for cmd, output in expected_outputs:
+            self.assertEqual(set(self.run_completer(make_parser(), cmd)), set(output))
+
+        expected_outputs = (
+            ("prog ", ["--foo", "--bar", "-h", "--help"]),
+            ("prog --b", ["--bar "])
+        )
+
+        for cmd, output in expected_outputs:
+            self.assertEqual(set(self.run_completer(make_parser(), cmd, print_suppressed=True)), set(output))
+
     def test_action_activation(self):
         def make_parser():
             parser = ArgumentParser()
@@ -218,7 +245,8 @@ class TestArgcomplete(unittest.TestCase):
             ("prog --url \"http://url1\" --email a", ["a@b.c", "a@b.d", "ab@c.d"]),
             ("prog --url \"http://url1\" --email \"a@", ['a@b.c', 'a@b.d']),
             ("prog --url \"http://url1\" --email \"a@b.c\" \"a@b.d\" \"a@", ['a@b.c', 'a@b.d']),
-            ("prog --url \"http://url1\" --email \"a@b.c\" \"a@b.d\" \"ab@c.d\" ", ["--url", "--email", "-h", "--help"]),
+            ("prog --url \"http://url1\" --email \"a@b.c\" \"a@b.d\" \"ab@c.d\" ",
+             ["--url", "--email", "-h", "--help"]),
         )
 
         for cmd, output in expected_outputs:
@@ -286,6 +314,7 @@ class TestArgcomplete(unittest.TestCase):
 
     def test_directory_completion(self):
         completer = DirectoriesCompleter()
+
         def c(prefix):
             return set(completer(prefix))
         with TempDir(prefix="test_dir", dir="."):
@@ -342,10 +371,11 @@ class TestArgcomplete(unittest.TestCase):
         expected_outputs = (
             ("prog ", ["--help", "eggs", "-h", "spam", "--age"]),
             ("prog --age 1 eggs", ["eggs "]),
-            ("prog --age 2 eggs ", [r"on\ a\ train", r"with\ a\ goat", r"on\ a\ boat", r"in\ the\ rain", "--help", "-h"]),
+            ("prog --age 2 eggs ",
+             [r"on\ a\ train", r"with\ a\ goat", r"on\ a\ boat", r"in\ the\ rain", "--help", "-h"]),
             ("prog eggs ", [r"on\ a\ train", r"with\ a\ goat", r"on\ a\ boat", r"in\ the\ rain", "--help", "-h"]),
             ("prog eggs \"on a", ['on a train', 'on a boat']),
-            ("prog eggs on\\ a", [r"on\ a\ train", "on\ a\ boat"]),
+            ("prog eggs on\\ a", [r"on\ a\ train", r"on\ a\ boat"]),
             ("prog spam ", ["iberico", "ham", "--help", "-h"]),
         )
 
@@ -357,21 +387,19 @@ class TestArgcomplete(unittest.TestCase):
 
     def test_non_ascii(self):
         def make_parser():
-            _str = ensure_bytes if USING_PYTHON2 else str
             parser = ArgumentParser()
-            # Python 2 argparse only works with byte strings or ascii unicode strings.
-            # Python 3 argparse only works with unicode strings.
-            parser.add_argument(_str("--книга"), choices=[
-                _str("Трудно быть богом"),
-                _str("Парень из преисподней"),
-                _str("Понедельник начинается в субботу"),
+            parser.add_argument("--книга", choices=[
+                "Трудно быть богом",
+                "Парень из преисподней",
+                "Понедельник начинается в субботу",
             ])
             return parser
 
         expected_outputs = (
             ("prog ", ["--книга", "-h", "--help"]),
-            ("prog --книга ", [r"Трудно\ быть\ богом", r"Парень\ из\ преисподней", r"Понедельник\ начинается\ в\ субботу"]),
-            ("prog --книга П", [r"Парень\ из\ преисподней", "Понедельник\ начинается\ в\ субботу"]),
+            ("prog --книга ",
+             [r"Трудно\ быть\ богом", r"Парень\ из\ преисподней", r"Понедельник\ начинается\ в\ субботу"]),
+            ("prog --книга П", [r"Парень\ из\ преисподней", r"Понедельник\ начинается\ в\ субботу"]),
             ("prog --книга Пу", [""]),
         )
 
@@ -416,7 +444,7 @@ class TestArgcomplete(unittest.TestCase):
         expected_outputs = (
             ("prog ", ["-h", "--help"], validators[0]),
             ("prog ", ["bus", "car", "-h", "--help"], validators[1]),
-            ("prog bu", ["bus "], validators[1]),
+            ("prog bu", ["bus", "car"], validators[1]),
             ("prog bus ", ["apple", "orange", "-h", "--help"], validators[1]),
             ("prog bus appl", ["apple "], validators[2]),
             ("prog bus cappl", [""], validators[2]),
@@ -477,6 +505,40 @@ class TestArgcomplete(unittest.TestCase):
         self.assertEqual("ttt", disp.get("--oh", ""))
         self.assertEqual("ccc", disp.get("--ch", ""))
 
+    def test_display_completions_with_aliases(self):
+        parser = ArgumentParser()
+        parser.add_subparsers().add_parser("a", aliases=["b", "c"], help="abc help")
+
+        # empty
+        completer = CompletionFinder(parser)
+        completer.rl_complete("", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"a b c": "abc help", "-h --help": "show this help message and exit"}, disp)
+
+        # a
+        completer = CompletionFinder(parser)
+        completer.rl_complete("a", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"a": "abc help", "": "show this help message and exit"}, disp)
+
+        # b
+        completer = CompletionFinder(parser)
+        completer.rl_complete("b", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"b": "abc help", "": "show this help message and exit"}, disp)
+
+        # c
+        completer = CompletionFinder(parser)
+        completer.rl_complete("c", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"c": "abc help", "": "show this help message and exit"}, disp)
+
+        # (
+        completer = CompletionFinder(parser)
+        completer.rl_complete("(", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"": "show this help message and exit"}, disp)
+
     def test_nargs_one_or_more(self):
         def make_parser():
             parser = ArgumentParser()
@@ -615,7 +677,10 @@ class TestArgcomplete(unittest.TestCase):
         expected_outputs = (
             ("prog ", {"long": long_opts, "short": short_opts, True: long_opts + short_opts, False: [""]}),
             ("prog --foo", {"long": ["--foo "], "short": ["--foo "], True: ["--foo "], False: ["--foo "]}),
-            ("prog --b", {"long": ["--bar", "--baz"], "short": ["--bar", "--baz"], True: ["--bar", "--baz"], False: ["--bar", "--baz"]}),
+            ("prog --b", {"long": ["--bar", "--baz"],
+                          "short": ["--bar", "--baz"],
+                          True: ["--bar", "--baz"],
+                          False: ["--bar", "--baz"]}),
             ("prog -z -x", {"long": ["-x "], "short": ["-x "], True: ["-x "], False: ["-x "]}),
         )
         for cmd, outputs in expected_outputs:
@@ -654,6 +719,7 @@ class TestArgcomplete(unittest.TestCase):
         expected_outputs = (
             ("prog ", ["--get", "--set", "name1", "name2"]),
             ("prog --", ["--get", "--set"]),
+            ("prog -- ", ["name1", "name2"]),
             ("prog --get ", ["--get", "name1", "name2"]),
             ("prog --get name1 ", ["--get "])
         )
@@ -711,6 +777,52 @@ class TestArgcomplete(unittest.TestCase):
         self.assertEqual(set(self.run_completer(make_parser(), 'prog -3 "')), {"\"'"})
         self.assertEqual(set(self.run_completer(make_parser(), "prog -3 '")), {"\"'"})
 
+    def test_shellcode_utility(self):
+        with NamedTemporaryFile() as fh:
+            sc = shellcode(["prog"], use_defaults=True, shell="bash", complete_arguments=None)
+            fh.write(sc.encode())
+            fh.flush()
+            subprocess.check_call(['bash', '-n', fh.name])
+        with NamedTemporaryFile() as fh:
+            sc = shellcode(["prog", "prog2"], use_defaults=False, shell="bash", complete_arguments=["-o", "nospace"])
+            fh.write(sc.encode())
+            fh.flush()
+            subprocess.check_call(['bash', '-n', fh.name])
+        with NamedTemporaryFile() as fh:
+            sc = shellcode(["prog"], use_defaults=True, shell="bash", complete_arguments=None,
+                           argcomplete_script="~/.bash_completion.d/prog.py")
+            fh.write(sc.encode())
+            fh.flush()
+            subprocess.check_call(['bash', '-n', fh.name])
+        sc = shellcode(["prog"], use_defaults=False, shell="tcsh", complete_arguments=["-o", "nospace"])
+        sc = shellcode(["prog"], use_defaults=False, shell="woosh", complete_arguments=["-o", "nospace"])
+        sc = shellcode(["prog"], shell="fish")
+        sc = shellcode(["prog"], shell="fish", argcomplete_script="~/.bash_completion.d/prog.py")
+
+    def test_option_help(self):
+        os.environ["_ARGCOMPLETE_DFS"] = "\t"
+        os.environ["_ARGCOMPLETE_SUPPRESS_SPACE"] = "1"
+        os.environ["_ARGCOMPLETE_SHELL"] = "fish"
+
+        p = ArgumentParser()
+        p.add_argument("--foo", help="foo" + IFS + "help")
+        p.add_argument("--bar", "--bar2", help="bar help")
+
+        subparsers = p.add_subparsers()
+        subparsers.add_parser("subcommand", help="subcommand help")
+        subparsers.add_parser("subcommand 2", help="subcommand 2 help")
+
+        completions = self.run_completer(p, "prog --f")
+        self.assertEqual(set(completions), {"--foo\tfoo help"})
+
+        completions = self.run_completer(p, "prog --b")
+        self.assertEqual(set(completions), {"--bar\tbar help", "--bar2\tbar help"})
+
+        completions = self.run_completer(p, "prog sub")
+        self.assertEqual(set(completions), {"subcommand\tsubcommand help", "subcommand 2\tsubcommand 2 help"})
+
+        os.environ["_ARGCOMPLETE_DFS"] = "invalid"
+        self.assertRaises(Exception, self.run_completer, p, "prog --b")
 
 class TestArgcompleteREPL(unittest.TestCase):
     def setUp(self):
@@ -845,8 +957,8 @@ class TestSplitLine(unittest.TestCase):
         self.assertEqual(self.prefix('a b c'), 'c')
 
     def test_escaped_special(self):
-        self.assertEqual(self.prefix('a\$b'), 'a$b')
-        self.assertEqual(self.prefix('a\`b'), 'a`b')
+        self.assertEqual(self.prefix(r'a\$b'), 'a$b')
+        self.assertEqual(self.prefix(r'a\`b'), 'a`b')
 
     def test_unescaped_special(self):
         self.assertEqual(self.prefix('a$b'), 'a$b')
@@ -854,8 +966,8 @@ class TestSplitLine(unittest.TestCase):
 
     @unittest.expectedFailure
     def test_escaped_special_in_double_quotes(self):
-        self.assertEqual(self.prefix('"a\$b'), 'a$b')
-        self.assertEqual(self.prefix('"a\`b'), 'a`b')
+        self.assertEqual(self.prefix(r'"a\$b'), 'a$b')
+        self.assertEqual(self.prefix(r'"a\`b'), 'a`b')
 
     def test_punctuation(self):
         self.assertEqual(self.prefix('a,'), 'a,')
@@ -872,6 +984,64 @@ class TestSplitLine(unittest.TestCase):
         self.assertEqual(self.wordbreak('"b:c=d" '), None)
 
 
+class TestCheckModule(unittest.TestCase):
+    def setUp(self):
+        self.dir = TempDir(prefix="test_dir_module", dir=".")
+        self.dir.__enter__()
+        # There is some odd bug that seems to only come up in Python 3.4 where
+        # using "." in sys.path sometimes won't find modules, so we'll use the
+        # full path each time.
+        sys.path.insert(0, os.getcwd())
+
+    def tearDown(self):
+        sys.path.pop(0)
+        self.dir.__exit__()
+
+    def test_module(self):
+        self._mkfile('module.py')
+        path = _check_module.find('module')
+        self.assertEqual(path, os.path.abspath('module.py'))
+        self.assertNotIn('module', sys.modules)
+
+    def test_package(self):
+        os.mkdir('package')
+        self._mkfile('package/__init__.py')
+        self._mkfile('package/module.py')
+        path = _check_module.find('package.module')
+        self.assertEqual(path, os.path.abspath('package/module.py'))
+        self.assertNotIn('package', sys.modules)
+        self.assertNotIn('package.module', sys.modules)
+
+    def test_subpackage(self):
+        os.mkdir('package')
+        self._mkfile('package/__init__.py')
+        os.mkdir('package/subpackage')
+        self._mkfile('package/subpackage/__init__.py')
+        self._mkfile('package/subpackage/module.py')
+        path = _check_module.find('package.subpackage.module')
+        self.assertEqual(path, os.path.abspath('package/subpackage/module.py'))
+        self.assertNotIn('package', sys.modules)
+        self.assertNotIn('package.subpackage', sys.modules)
+        self.assertNotIn('package.subpackage.module', sys.modules)
+
+    def test_package_main(self):
+        os.mkdir('package')
+        self._mkfile('package/__init__.py')
+        self._mkfile('package/__main__.py')
+        path = _check_module.find('package')
+        self.assertEqual(path, os.path.abspath('package/__main__.py'))
+        self.assertNotIn('package', sys.modules)
+
+    def test_not_package(self):
+        self._mkfile('module.py')
+        with self.assertRaisesRegex(Exception, 'module is not a package'):
+            _check_module.find('module.bad')
+        self.assertNotIn('module', sys.modules)
+
+    def _mkfile(self, path):
+        open(path, 'w').close()
+
+
 class _TestSh(object):
     """
     Contains tests which should work in any shell using argcomplete.
@@ -894,15 +1064,24 @@ class _TestSh(object):
     """
     sh = None
     expected_failures = []
+    skipped = []
 
     @classmethod
     def setUpClass(cls, *args, **kwargs):
         for name in cls.expected_failures:
             test = getattr(cls, name)
+
             @unittest.expectedFailure
             def wrapped(self, test=test):
                 test(self)
             setattr(cls, name, wrapped)
+        for name in cls.skipped:
+            test = getattr(cls, name)
+
+            @unittest.skip('skip')
+            def wrapped(self, test=test):
+                pass
+            setattr(cls, name, wrapped)
         super(_TestSh, cls).setUpClass(*args, **kwargs)
 
     def setUp(self):
@@ -990,6 +1169,14 @@ class _TestSh(object):
     def test_completion_environment(self):
         self.assertEqual(self.sh.run_command('prog env o\t'), 'ok\r\n')
 
+    def test_comp_point(self):
+        # Use environment variable to change how prog behaves
+        self.assertEqual(self.sh.run_command('export POINT=1'), '')
+        self.assertEqual(self.sh.run_command('prog point hi\t'), '13\r\n')
+        self.assertEqual(self.sh.run_command('prog point hi \t'), '14\r\n')
+        self.assertEqual(self.sh.run_command('prog point 你好嘚瑟\t'), '15\r\n')
+        self.assertEqual(self.sh.run_command('prog point 你好嘚瑟 \t'), '16\r\n')
+
 
 class TestBash(_TestSh, unittest.TestCase):
     expected_failures = [
@@ -1000,15 +1187,23 @@ class TestBash(_TestSh, unittest.TestCase):
         # This requires compopt which is not available in 3.x.
         expected_failures.append('test_quoted_exact')
 
-    install_cmd = 'eval "$(register-python-argcomplete prog)"'
+    # 'dummy' argument unused; checks multi-command registration works
+    # by passing 'prog' as the second argument.
+    install_cmd = 'eval "$(register-python-argcomplete dummy prog)"'
 
     def setUp(self):
         sh = pexpect.replwrap.bash()
         path = ':'.join([os.path.join(BASE_DIR, 'scripts'), TEST_DIR, '$PATH'])
         sh.run_command('export PATH={0}'.format(path))
         sh.run_command('export PYTHONPATH={0}'.format(BASE_DIR))
+        # Disable the "python" module provided by bash-completion
+        sh.run_command('complete -r python python2 python3')
         output = sh.run_command(self.install_cmd)
         self.assertEqual(output, '')
+        # Register a dummy completion with an external argcomplete script
+        # to ensure this doesn't overwrite our previous registration.
+        output = sh.run_command('eval "$(register-python-argcomplete dummy --external-argcomplete-script dummy)"')
+        self.assertEqual(output, '')
         self.sh = sh
 
     def test_one_space_after_exact(self):
@@ -1017,6 +1212,26 @@ class TestBash(_TestSh, unittest.TestCase):
         result = self.sh.run_command('prog basic f\t"\1echo "')
         self.assertEqual(result, 'prog basic foo \r\n')
 
+    def test_debug_output(self):
+        self.assertEqual(self.sh.run_command('prog debug f\t'), 'foo\r\n')
+        self.sh.run_command('export _ARC_DEBUG=1')
+        output = self.sh.run_command('prog debug f\t')
+        self.assertIn('PYTHON_ARGCOMPLETE_STDOUT\r\n', output)
+        self.assertIn('PYTHON_ARGCOMPLETE_STDERR\r\n', output)
+        self.assertTrue(output.endswith('foo\r\n'))
+
+    def test_temp_file(self):
+        self.sh.run_command('export ARGCOMPLETE_USE_TEMPFILES=1')
+        self.assertEqual(self.sh.run_command('prog basic f\t'), 'foo\r\n')
+        # Confirm we used a temp file by searching for the debug message.
+        self.sh.run_command('export _ARC_DEBUG=1')
+        output = self.sh.run_command('prog basic f\t')
+        self.assertIn('Using output file ', output)
+
+    def test_nounset(self):
+        self.sh.run_command('set -o nounset')
+        self.test_simple_completion()
+
 
 @unittest.skipIf(BASH_MAJOR_VERSION < 4, 'complete -D not supported')
 class TestBashGlobal(TestBash):
@@ -1024,11 +1239,11 @@ class TestBashGlobal(TestBash):
 
     def test_python_completion(self):
         self.sh.run_command('cd ' + TEST_DIR)
-        self.assertEqual(self.sh.run_command('python ./prog basic f\t'), 'foo\r\n')
+        self.assertEqual(self.sh.run_command('python3 ./prog basic f\t'), 'foo\r\n')
 
     def test_python_filename_completion(self):
         self.sh.run_command('cd ' + TEST_DIR)
-        self.assertEqual(self.sh.run_command('python ./pro\tbasic f\t'), 'foo\r\n')
+        self.assertEqual(self.sh.run_command('python3 ./pro\tbasic f\t'), 'foo\r\n')
 
     def test_python_not_executable(self):
         """Test completing a script that cannot be run directly."""
@@ -1037,8 +1252,60 @@ class TestBashGlobal(TestBash):
             shutil.copy(prog, '.')
             self.sh.run_command('cd ' + os.getcwd())
             self.sh.run_command('chmod -x ./prog')
-            self.assertIn('Permission denied', self.sh.run_command('./prog'))
-            self.assertEqual(self.sh.run_command('python ./prog basic f\t'), 'foo\r\n')
+            # Ensure prog is no longer able to be run as "./prog".
+            self.assertIn('<<126>>', self.sh.run_command('./prog; echo "<<$?>>"'))
+            # Ensure completion still functions when run via python.
+            self.assertEqual(self.sh.run_command('python3 ./prog basic f\t'), 'foo\r\n')
+
+    def test_python_module(self):
+        """Test completing a module run with python -m."""
+        prog = os.path.join(TEST_DIR, 'prog')
+        with TempDir(prefix='test_dir_py', dir='.'):
+            os.mkdir('package')
+            open('package/__init__.py', 'w').close()
+            shutil.copy(prog, 'package/prog.py')
+            self.sh.run_command('cd ' + os.getcwd())
+            self.assertEqual(self.sh.run_command('python3 -m package.prog basic f\t'), 'foo\r\n')
+
+    def _test_console_script(self, package=False, wheel=False):
+        with TempDir(prefix='test_dir_py', dir='.'):
+            self.sh.run_command('cd ' + os.getcwd())
+            self.sh.run_command('export PATH=$PATH:./bin')
+            self.sh.run_command('export PYTHONPATH=.:$PYTHONPATH')
+            test_package = os.path.join(TEST_DIR, 'test_package')
+            command = 'pip install {} --target .'.format(test_package)
+            if not wheel:
+                command += ' --no-binary :all:'
+                if sys.platform == 'darwin':
+                    # Work around https://stackoverflow.com/questions/24257803
+                    command += ' --install-option="--prefix="'
+            install_output = self.sh.run_command(command)
+            self.assertEqual(self.sh.run_command('echo $?'), '0\r\n', install_output)
+            command = 'test-module'
+            if package:
+                command = 'test-package'
+            command += ' a\t'
+            self.assertEqual(self.sh.run_command(command), 'arg\r\n')
+
+    @unittest.skipIf(os.uname()[0] == "Darwin", "Skip test that fails on MacOS")
+    def test_console_script_module(self):
+        """Test completing a console_script for a module."""
+        self._test_console_script()
+
+    @unittest.skipIf(os.uname()[0] == "Darwin", "Skip test that fails on MacOS")
+    def test_console_script_package(self):
+        """Test completing a console_script for a package."""
+        self._test_console_script(package=True)
+
+    @unittest.skipIf(os.uname()[0] == "Darwin", "Skip test that fails on MacOS")
+    def test_console_script_module_wheel(self):
+        """Test completing a console_script for a module from a wheel."""
+        self._test_console_script(wheel=True)
+
+    @unittest.skipIf(os.uname()[0] == "Darwin", "Skip test that fails on MacOS")
+    def test_console_script_package_wheel(self):
+        """Test completing a console_script for a package from a wheel."""
+        self._test_console_script(package=True, wheel=True)
 
 
 class TestTcsh(_TestSh, unittest.TestCase):
@@ -1048,6 +1315,8 @@ class TestTcsh(_TestSh, unittest.TestCase):
         'test_continuation',
         'test_parse_special_characters',
         'test_parse_special_characters_dollar',
+        # Test case doesn't work under tcsh, could be fixed.
+        'test_comp_point',
     ]
 
     def setUp(self):
@@ -1055,7 +1324,13 @@ class TestTcsh(_TestSh, unittest.TestCase):
         path = ' '.join([os.path.join(BASE_DIR, 'scripts'), TEST_DIR, '$path'])
         sh.run_command('set path = ({0})'.format(path))
         sh.run_command('setenv PYTHONPATH {0}'.format(BASE_DIR))
-        output = sh.run_command('eval `register-python-argcomplete --shell tcsh prog`')
+        # 'dummy' argument unused; checks multi-command registration works
+        # by passing 'prog' as the second argument.
+        output = sh.run_command('eval `register-python-argcomplete --shell tcsh dummy prog`')
+        self.assertEqual(output, '')
+        # Register a dummy completion with an external argcomplete script
+        # to ensure this doesn't overwrite our previous registration.
+        output = sh.run_command('eval `register-python-argcomplete --shell tcsh dummy --external-argcomplete-script dummy`')
         self.assertEqual(output, '')
         self.sh = sh
 
@@ -1067,6 +1342,69 @@ class TestTcsh(_TestSh, unittest.TestCase):
             self.sh.run_command('')
 
 
+class TestFish(_TestSh, unittest.TestCase):
+    expected_failures = [
+        'test_parse_special_characters',
+        'test_comp_point',
+    ]
+    if FISH_VERSION_TUPLE < (3, 1):
+        expected_failures.extend([
+            'test_special_characters_double_quoted'
+        ])
+
+    skipped = [
+        'test_single_quotes_in_single_quotes',
+        'test_parse_special_characters_dollar'
+    ]
+
+    def setUp(self):
+
+        sh = Shell('fish')
+        path = ' '.join([os.path.join(BASE_DIR, 'scripts'), TEST_DIR, '$PATH'])
+        sh.run_command('set -x PATH {0}'.format(path))
+        sh.run_command('set -x PYTHONPATH {0}'.format(BASE_DIR))
+        # 'dummy' argument unused; checks multi-command registration works
+        # by passing 'prog' as the second argument.
+        output = sh.run_command('register-python-argcomplete --shell fish dummy prog | source')
+        self.assertEqual(output, '')
+        # Register a dummy completion with an external argcomplete script
+        # to ensure this doesn't overwrite our previous registration.
+        output = sh.run_command('register-python-argcomplete --shell fish dummy --external-argcomplete-script dummy | source')
+        self.assertEqual(output, '')
+        self.sh = sh
+
+    def tearDown(self):
+        # The shell wrapper is fragile; exactly which exception is raised
+        # differs depending on environment.
+        with self.assertRaises((pexpect.EOF, OSError)):
+            self.sh.run_command('exit')
+            self.sh.run_command('')
+
+
+@unittest.skipIf(FISH_VERSION_TUPLE < (3, 3), "Path completion is fixed in fish 3.3")
+class TestFishPathCompletion(unittest.TestCase):
+    def setUp(self):
+        sh = Shell('fish')
+        path = ' '.join([os.path.join(BASE_DIR, 'scripts'), '$PATH'])
+        sh.run_command('set -x PATH {0}'.format(path))
+        sh.run_command('set -x PYTHONPATH {0}'.format(BASE_DIR))
+        self.assertEqual(sh.run_command('register-python-argcomplete --shell fish {0}/prog | source'.format(TEST_DIR)), '')
+
+        self.sh = sh
+
+    def test_path_completion(self):
+        self.assertEqual(self.sh.run_command('{}/prog basic f\t'.format(TEST_DIR)), 'foo\r\n')
+        self.sh.run_command('cd {}'.format(TEST_DIR))
+        self.assertEqual(self.sh.run_command('./prog basic f\t'), 'foo\r\n')
+
+    def tearDown(self):
+        # The shell wrapper is fragile; exactly which exception is raised
+        # differs depending on environment.
+        with self.assertRaises((pexpect.EOF, OSError)):
+            self.sh.run_command('exit')
+            self.sh.run_command('')
+
+
 class Shell(object):
     def __init__(self, shell):
         self.child = pexpect.spawn(shell, encoding='utf-8')
@@ -1074,13 +1412,31 @@ class Shell(object):
     def run_command(self, command):
         try:
             self.child.sendline(r"echo -n \#\#\#; {0}; echo -n \#\#\#".format(command))
-            self.child.expect_exact('###', timeout=1)
-            self.child.expect_exact('###', timeout=1)
+            self.child.expect_exact('###', timeout=5)
+            self.child.expect_exact('###', timeout=5)
             return self.child.before
         finally:
             # Send Ctrl+C in case we get stuck.
             self.child.sendline('\x03')
 
 
+class Warn(unittest.TestCase):
+    def test_warn(self):
+        @contextlib.contextmanager
+        def redirect_debug_stream(stream):
+            debug_stream = argcomplete.debug_stream
+            argcomplete.debug_stream = stream
+            try:
+                yield
+            finally:
+                argcomplete.debug_stream = debug_stream
+
+        test_stream = StringIO()
+        with redirect_debug_stream(test_stream):
+            warn("My hands are tied")
+
+        self.assertEqual("\nMy hands are tied\n", test_stream.getvalue())
+
+
 if __name__ == "__main__":
     unittest.main()
